
api = "alsa"
bend_range = 2 # semitones
# Midi input Port Identifier
synths = ["fluid"]
# number of channels (max 16 client ports X 16 chnls = 256 for fluidsynth?)
# max ports for fluid synth is 8
port_count = 1
